#pragma OPENCL EXTENSION cl_khr_fp64 : enable
#define ln2 0.6931471805599453

#include "mandelbrot.h"

//the actual mandelbrot function
double mandelbrot(double c_r, double c_i, double z0_r, double z0_i, double max_steps)
{
	complex_double f, c, tmp;
	double k, ret;

    c.r = c_r; c.i = c_i;
    f.r = z0_r; f.i = z0_i;

	k = 0;
	while (sqrt(f.r*f.r + f.i*f.i) <= escape)
	{
		tmp.r = (f.r*f.r) - (f.i*f.i);
		tmp.i = (f.r*f.i) + (f.i*f.r);

		f.r = tmp.r;
		f.i = tmp.i;

		f.r += c.r;
		f.i += c.i;

		k += 1;

		if (k > max_steps)
		{
	        return 0.0;
		}
	}

    ret = log(f.r*f.r + f.i*f.i) / 2;
    ret = log(ret / ln2) / ln2;
    ret = log(k+1-ret);

    return (ret > 0) ? ret : 0;
}

//compute value for pixel
__kernel void kern(double __global *img, double __global *params)
{
    double x,y;
	complex_double f, c;
    int i = get_global_id(0);
    int N = get_local_size(0);
    for(; i < (X_DIM*Y_DIM); i+=N)
    {
        x = i % X_DIM;
        y = i / X_DIM;
        //normalize position to -1/scale ... 1/scale
        x = ((x/X_DIM) * 2 - 1) / params[params_scale];
        y = ((y/Y_DIM) * 2 - 1) / params[params_scale];

        c.r = params[params_c_r] + x*params[params_x_c_r] + y*params[params_y_c_r];
        c.i = params[params_c_i] + x*params[params_x_c_i] + y*params[params_y_c_i];
        f.r = params[params_z0_r] + x*params[params_x_z0_r] + y*params[params_y_z0_r];
        f.i = params[params_z0_i] + x*params[params_x_z0_i] + y*params[params_y_z0_i];

        img[i] = mandelbrot(c.r, c.i, f.r, f.i, params[params_maxsteps]);
    }
}
